@@ -29,7 +29,7 @@
 CORBA_Object
 cspi_dup_ref (CORBA_Object object)
 {
-  return bonobo_object_dup_ref (object, NULL);
+  return bonobo_object_dup_ref (object, cspi_ev ());
 }
 
 void
@@ -41,7 +41,7 @@ cspi_release_unref (CORBA_Object object)
 SPIBoolean
 cspi_check_ev (const char *error_string)
 {
-  CORBA_Environment *ev = cspi_ev ();
+  CORBA_Environment *ev = cspi_peek_ev ();
 
   if (ev->_major != CORBA_NO_EXCEPTION)
     {
@@ -49,8 +49,10 @@ cspi_check_ev (const char *error_string)
 
       err = bonobo_exception_get_text (ev);
 
-      fprintf (stderr, "Warning: AT-SPI error: %s: %s\n",
-	       error_string, err);
+      if (!_cspi_exception_throw (ev, error_string)) {
+	fprintf (stderr, "Warning: AT-SPI error: %s: %s\n",
+		 error_string, err);
+      }
 
       g_free (err);
 
