@@ -2,7 +2,7 @@
  *
  * (Gnome Accessibility Project; http://developer.gnome.org/projects/gap)
  *
- * Copyright 2001, 2002 Sun Microsystems Inc.,
+ * Copyright 2001, 2003 Sun Microsystems Inc.,
  * Copyright 2001, 2002 Ximian, Inc.
  *
  * This library is free software; you can redistribute it and/or
@@ -32,6 +32,7 @@
 #include <string.h>
 #include <ctype.h>
 #include <stdio.h>
+#include <sys/time.h>
 #include <bonobo/bonobo-exception.h>
 
 #include <X11/Xlib.h>
@@ -39,6 +40,13 @@
 #include <X11/XKBlib.h>
 #define XK_MISCELLANY
 #include <X11/keysymdef.h>
+
+#ifdef HAVE_XEVIE
+#include <X11/Xproto.h>
+#include <X11/X.h>
+#include <X11/extensions/Xevie.h>
+#endif /* HAVE_XEVIE */
+
 #include <gdk/gdk.h>
 #include <gdk/gdkx.h> /* TODO: hide dependency (wrap in single porting file) */
 #include <gdk/gdkkeysyms.h>
@@ -59,7 +67,8 @@ static unsigned int mouse_mask_state = 0;
 static unsigned int mouse_button_mask =
   Button1Mask | Button2Mask | Button3Mask | Button4Mask | Button5Mask;
 static unsigned int key_modifier_mask =
-  Mod1Mask | Mod2Mask | Mod3Mask | Mod4Mask | Mod5Mask | ShiftMask | LockMask | ControlMask;
+  Mod1Mask | Mod2Mask | Mod3Mask | Mod4Mask | Mod5Mask | ShiftMask | LockMask | ControlMask | SPI_KEYMASK_NUMLOCK;
+static unsigned int _numlock_physical_mask = Mod2Mask; /* a guess, will be reset */
 
 static GQuark spi_dec_private_quark = 0;
 
@@ -114,7 +123,7 @@ static gboolean spi_controller_update_key_grabs               (SpiDEController
 static gboolean spi_controller_register_device_listener       (SpiDEController           *controller,
 							       DEControllerListener      *l,
 							       CORBA_Environment         *ev);
-static void     spi_device_event_controller_forward_key_event (SpiDEController           *controller,
+static gboolean spi_device_event_controller_forward_key_event (SpiDEController           *controller,
 							       const XEvent              *event);
 static void     spi_deregister_controller_device_listener (SpiDEController            *controller,
 					                   DEControllerListener *listener,
@@ -137,10 +146,60 @@ static gboolean spi_dec_poll_mouse_idle (gpointer data);
 
 /* Private methods */
 
-static KeyCode
-keycode_for_keysym (long keysym)
+static unsigned int
+keysym_mod_mask (KeySym keysym, KeyCode keycode)
 {
-  return XKeysymToKeycode (spi_get_display (), (KeySym) keysym);
+	/* we really should use XKB and look directly at the keymap */
+	/* this is very inelegant */
+	Display *display = spi_get_display ();
+	unsigned int mods_rtn = 0;
+	unsigned int retval = 0;
+	KeySym sym_rtn;
+
+	if (XkbLookupKeySym (display, keycode, 0, &mods_rtn, &sym_rtn) &&
+	    (sym_rtn == keysym)) {
+		retval = 0;
+	}
+	else if (XkbLookupKeySym (display, keycode, ShiftMask, &mods_rtn, &sym_rtn) &&
+		 (sym_rtn == keysym)) {
+		retval = ShiftMask;
+	}
+	else if (XkbLookupKeySym (display, keycode, Mod2Mask, &mods_rtn, &sym_rtn) &&
+		 (sym_rtn == keysym)) {
+		retval = Mod2Mask;
+	}
+	else if (XkbLookupKeySym (display, keycode, Mod3Mask, &mods_rtn, &sym_rtn) &&
+		 (sym_rtn == keysym)) {
+		retval = Mod3Mask;
+	}
+	else if (XkbLookupKeySym (display, keycode, 
+				  ShiftMask | Mod2Mask, &mods_rtn, &sym_rtn) &&
+		 (sym_rtn == keysym)) {
+		retval = (Mod2Mask | ShiftMask);
+	}
+	else if (XkbLookupKeySym (display, keycode, 
+				  ShiftMask | Mod3Mask, &mods_rtn, &sym_rtn) &&
+		 (sym_rtn == keysym)) {
+		retval = (Mod3Mask | ShiftMask);
+	}
+	else if (XkbLookupKeySym (display, keycode, 
+				  ShiftMask | Mod4Mask, &mods_rtn, &sym_rtn) &&
+		 (sym_rtn == keysym)) {
+		retval = (Mod4Mask | ShiftMask);
+	}
+	else
+		retval = 0xFFFF;
+	return retval;
+}
+
+static KeyCode
+keycode_for_keysym (long keysym, unsigned int *modmask)
+{
+	KeyCode keycode = 0;
+	keycode = XKeysymToKeycode (spi_get_display (), (KeySym) keysym);
+	if (modmask) 
+		*modmask = keysym_mod_mask (keysym, keycode);
+	return keycode;
 }
 
 static DEControllerGrabMask *
@@ -185,13 +244,20 @@ spi_dec_set_unlatch_pending (SpiDEController *controller, unsigned mask)
   DEControllerPrivateData *priv = 
     g_object_get_qdata (G_OBJECT (controller), spi_dec_private_quark);
 #ifdef SPI_XKB_DEBUG
-  if (priv->xkb_latch_mask) fprintf (stderr, "unlatch pending! %x\n", mask);
+  if (priv->xkb_latch_mask) fprintf (stderr, "unlatch pending! %x\n", 
+				     priv->xkb_latch_mask);
 #endif
   priv->pending_xkb_mod_relatch_mask |= priv->xkb_latch_mask; 
 }
  
-static gint poll_count = 0;
-
+static void
+spi_dec_clear_unlatch_pending (SpiDEController *controller)
+{
+  DEControllerPrivateData *priv = 
+    g_object_get_qdata (G_OBJECT (controller), spi_dec_private_quark);
+  priv->xkb_latch_mask = 0; 
+}
+ 
 static gboolean
 spi_dec_button_update_and_emit (SpiDEController *controller, 
 				guint mask_return)
@@ -280,7 +346,7 @@ spi_dec_button_update_and_emit (SpiDEController *controller,
 #endif
 	snprintf (event_name, 22, "mouse:button:%d%c", button_number,
 		  (is_down) ? 'p' : 'r');
-	/* TODO: distinguish between physical and 
+	/* TODO: FIXME distinguish between physical and 
 	 * logical buttons 
 	 */
 	mouse_e.type      = (is_down) ? 
@@ -327,7 +393,6 @@ spi_dec_mouse_check (SpiDEController *controller,
   Accessibility_Event e;
   CORBA_Environment ev;
   int win_x_return,win_y_return;
-  int poll_count_modulus = 10;
   unsigned int mask_return;
   Window root_return, child_return;
   Display *display = spi_get_display ();
@@ -349,20 +414,17 @@ spi_dec_mouse_check (SpiDEController *controller,
       while (spi_dec_button_update_and_emit (controller, mask_return));
     }
 
-  if (poll_count++ == poll_count_modulus) {
-    poll_count = 0;
-    e.type = "mouse:abs";  
-    e.source = BONOBO_OBJREF (controller->registry->desktop);
-    e.detail1 = *x;
-    e.detail2 = *y;
-    spi_init_any_nil (&e.any_data);
-    CORBA_exception_init (&ev);
-    Accessibility_Registry_notifyEvent (BONOBO_OBJREF (controller->registry),
-					&e,
-					&ev);
-  }
   if (*x != last_mouse_pos->x || *y != last_mouse_pos->y) 
     {
+      e.type = "mouse:abs";  
+      e.source = BONOBO_OBJREF (controller->registry->desktop);
+      e.detail1 = *x;
+      e.detail2 = *y;
+      spi_init_any_nil (&e.any_data);
+      CORBA_exception_init (&ev);
+      Accessibility_Registry_notifyEvent (BONOBO_OBJREF (controller->registry),
+					  &e,
+					  &ev);
       e.type = "mouse:rel";  
       e.source = BONOBO_OBJREF (controller->registry->desktop);
       e.detail1 = *x - last_mouse_pos->x;
@@ -395,6 +457,13 @@ spi_dec_emit_modifier_event (SpiDEController *controller, guint prev_mask,
   fprintf (stderr, "MODIFIER CHANGE EVENT! %x to %x\n", 
 	   prev_mask, current_mask);
 #endif
+
+  /* set bits for the virtual modifiers like NUMLOCK */
+  if (prev_mask & _numlock_physical_mask) 
+    prev_mask |= SPI_KEYMASK_NUMLOCK;
+  if (current_mask & _numlock_physical_mask) 
+    current_mask |= SPI_KEYMASK_NUMLOCK;
+
   e.type = "keyboard:modifiers";  
   e.source = BONOBO_OBJREF (controller->registry->desktop);
   e.detail1 = prev_mask & key_modifier_mask;
@@ -451,6 +520,7 @@ spi_dec_poll_mouse_moving (gpointer data)
     }
 }
 
+#ifdef WE_NEED_UGRAB_MOUSE
 static int
 spi_dec_ungrab_mouse (gpointer data)
 {
@@ -462,6 +532,7 @@ spi_dec_ungrab_mouse (gpointer data)
 	  }
 	return FALSE;
 }
+#endif
 
 static void
 spi_dec_init_mouse_listener (SpiRegistry *registry)
@@ -471,10 +542,15 @@ spi_dec_init_mouse_listener (SpiRegistry *registry)
 
   if (display)
     {
-      XGrabButton (display, AnyButton, AnyModifier,
-		   gdk_x11_get_default_root_xwindow (),
-		   True, ButtonPressMask | ButtonReleaseMask,
-		   GrabModeSync, GrabModeAsync, None, None);
+      if (XGrabButton (display, AnyButton, AnyModifier,
+		       gdk_x11_get_default_root_xwindow (),
+		       True, ButtonPressMask | ButtonReleaseMask,
+		       GrabModeSync, GrabModeAsync, None, None) != Success) {
+#ifdef SPI_DEBUG
+	fprintf (stderr, "WARNING: could not grab mouse buttons!\n");
+#endif
+	;
+      }
       XSync (display, False);
 #ifdef SPI_DEBUG
       fprintf (stderr, "mouse buttons grabbed\n");
@@ -482,6 +558,24 @@ spi_dec_init_mouse_listener (SpiRegistry *registry)
     }
 }
 
+/**
+ * Eventually we can use this to make the marshalling of mask types
+ * more sane, but for now we just use this to detect 
+ * the use of 'virtual' masks such as Mumlock and convert them to
+ * system-specific mask values (i.e. ModMask).
+ * 
+ **/
+static Accessibility_ControllerEventMask
+spi_dec_translate_mask (Accessibility_ControllerEventMask mask)
+{
+  DEControllerPrivateData *priv;
+
+  if (mask == SPI_KEYMASK_NUMLOCK) {
+    mask = _numlock_physical_mask;
+  }
+  return mask;
+}
+
 static DEControllerKeyListener *
 spi_dec_key_listener_new (CORBA_Object                            l,
 			  const Accessibility_KeySet             *keys,
@@ -494,7 +588,7 @@ spi_dec_key_listener_new (CORBA_Object                            l,
   key_listener->listener.object = bonobo_object_dup_ref (l, ev);
   key_listener->listener.type = SPI_DEVICE_TYPE_KBD;
   key_listener->keys = ORBit_copy_value (keys, TC_Accessibility_KeySet);
-  key_listener->mask = mask;
+  key_listener->mask = spi_dec_translate_mask (mask);
   key_listener->listener.typeseq = ORBit_copy_value (typeseq, TC_Accessibility_EventTypeSeq);
   if (mode)
     key_listener->mode = ORBit_copy_value (mode, TC_Accessibility_EventListenerMode);
@@ -684,7 +778,10 @@ spi_controller_register_global_keygrabs (SpiDEController         *controller,
 					 DEControllerKeyListener *key_listener)
 {
   handle_keygrab (controller, key_listener, _register_keygrab);
-  return spi_controller_update_key_grabs (controller, NULL);
+  if (controller->xevie_display == NULL)
+    return spi_controller_update_key_grabs (controller, NULL);
+  else
+    return TRUE;
 }
 
 static void
@@ -692,7 +789,8 @@ spi_controller_deregister_global_keygrabs (SpiDEController         *controller,
 					   DEControllerKeyListener *key_listener)
 {
   handle_keygrab (controller, key_listener, _deregister_keygrab);
-  spi_controller_update_key_grabs (controller, NULL);
+  if (controller->xevie_display == NULL)
+    spi_controller_update_key_grabs (controller, NULL);
 }
 
 static gboolean
@@ -734,6 +832,7 @@ spi_controller_notify_mouselisteners (SpiDEController                 *controlle
   GSList  *notify = NULL, *l2;
   GList  **listeners = &controller->mouse_listeners;
   gboolean is_consumed;
+  gboolean found = FALSE;
 
   if (!listeners)
     {
@@ -753,12 +852,13 @@ spi_controller_notify_mouselisteners (SpiDEController                 *controlle
 	       /* we clone (don't dup) the listener, to avoid refcount inc. */
 	       notify = g_slist_prepend (notify,
 					 spi_listener_clone (listener, ev));
+               found = TRUE;
 	     }
          }
     }
 
 #ifdef SPI_KEYEVENT_DEBUG
-  if (!notify)
+  if (!found)
     {
       g_print ("no match for event\n");
     }
@@ -814,7 +914,7 @@ spi_device_event_controller_forward_mouse_event (SpiDEController *controller,
   int button = xbutton_event->button;
   
   unsigned int mouse_button_state = xbutton_event->state;
-  
+
   switch (button)
     {
     case 1:
@@ -846,7 +946,7 @@ spi_device_event_controller_forward_mouse_event (SpiDEController *controller,
   snprintf (event_name, 22, "mouse:button:%d%c", button,
 	    (xevent->type == ButtonPress) ? 'p' : 'r');
 
-  /* TODO: distinguish between physical and logical buttons */
+  /* TODO: FIXME distinguish between physical and logical buttons */
   mouse_e.type      = (xevent->type == ButtonPress) ? 
                       Accessibility_BUTTON_PRESSED_EVENT :
                       Accessibility_BUTTON_RELEASED_EVENT;
@@ -884,7 +984,7 @@ spi_device_event_controller_forward_mouse_event (SpiDEController *controller,
   /* if client wants to consume this event, and XKB latch state was
    *   unset by this button event, we reset it
    */
-  if (is_consumed && (xkb_mod_unlatch_occurred))
+  if (is_consumed && xkb_mod_unlatch_occurred)
     spi_dec_set_unlatch_pending (controller, mouse_mask_state);
   
   XAllowEvents (spi_get_display (),
@@ -905,7 +1005,17 @@ global_filter_fn (GdkXEvent *gdk_xevent, GdkEvent *event, gpointer data)
 
   if (xevent->type == KeyPress || xevent->type == KeyRelease)
     {
-      spi_device_event_controller_forward_key_event (controller, xevent);
+      if (controller->xevie_display == NULL)
+        {
+          gboolean is_consumed =
+            spi_device_event_controller_forward_key_event (controller, xevent);
+
+          if (is_consumed)
+            XAllowEvents (spi_get_display (), AsyncKeyboard, CurrentTime);
+          else
+            XAllowEvents (spi_get_display (), ReplayKeyboard, CurrentTime);
+        }
+
       return GDK_FILTER_CONTINUE;
     }
   if (xevent->type == ButtonPress || xevent->type == ButtonRelease)
@@ -953,6 +1063,7 @@ global_filter_fn (GdkXEvent *gdk_xevent, GdkEvent *event, gpointer data)
 			      &= ~(XkbAX_StickyKeysFBMask);
 	        XkbChangeControls (display, priv->xkb_desc, &changes);
 	      }
+	      /* TODO: account for lock as well as latch */
 	      XkbLatchModifiers (display,
 				 XkbUseCoreKbd,
 				 priv->pending_xkb_mod_relatch_mask,
@@ -1005,7 +1116,6 @@ spi_controller_register_with_devices (SpiDEController *controller)
 	  g_object_get_qdata (G_OBJECT (controller), spi_dec_private_quark);	 
   /* FIXME: should check for extension first! */
   XTestGrabControl (spi_get_display (), True);
-  priv->xkb_desc = XkbGetMap (spi_get_display (), 0, XkbUseCoreKbd);
 
   /* calls to device-specific implementations and routines go here */
   /* register with: keyboard hardware code handler */
@@ -1017,10 +1127,14 @@ spi_controller_register_with_devices (SpiDEController *controller)
 				      &priv->xkb_base_error_code, NULL, NULL);
   if (priv->have_xkb)
     {
+      priv->xkb_desc = XkbGetMap (spi_get_display (), 0, XkbUseCoreKbd);
       XkbSelectEvents (spi_get_display (),
 		       XkbUseCoreKbd,
 		       XkbStateNotifyMask, XkbStateNotifyMask);	    
+      _numlock_physical_mask = XkbKeysymToModifiers (spi_get_display (), 
+						     XK_Num_Lock);
     }	
+
   gdk_window_add_filter (NULL, global_filter_fn, controller);
 
   gdk_window_set_events (gdk_get_default_root_window (),
@@ -1053,10 +1167,13 @@ spi_key_set_contains_key (Accessibility_KeySet            *key_set,
   for (i = 0; i < len; ++i)
     {
 #ifdef SPI_KEYEVENT_DEBUG	    
-      g_print ("key_set[%d] = %d; key_event %d, code %d, string %s\n",
-	        i, (int) key_set->_buffer[i].keycode,
-	       (int) key_event->id, (int) key_event->hw_code,
-	       key_event->event_string); 
+      g_print ("key_set[%d] event = %d, code = %d; key_event %d, code %d, string %s\n",
+                i,
+                (int)key_set->_buffer[i].keysym,
+                (int) key_set->_buffer[i].keycode,
+                (int) key_event->id,
+                (int) key_event->hw_code,
+                key_event->event_string); 
 #endif
       if (key_set->_buffer[i].keysym == (CORBA_long) key_event->id)
         {
@@ -1072,7 +1189,7 @@ spi_key_set_contains_key (Accessibility_KeySet            *key_set,
           return TRUE;
 	}
     }
-  
+
   return FALSE;
 }
 
@@ -1117,7 +1234,7 @@ spi_key_event_matches_listener (const Accessibility_DeviceEvent *key_event,
 				DEControllerKeyListener         *listener,
 				CORBA_boolean                    is_system_global)
 {
-  if ((key_event->modifiers == (CORBA_unsigned_short) (listener->mask & 0xFFFF)) &&
+  if ((key_event->modifiers == (CORBA_unsigned_short) (listener->mask & 0xFF)) &&
        spi_key_set_contains_key (listener->keys, key_event) &&
        spi_eventtype_seq_contains_event (listener->listener.typeseq, key_event) && 
       (is_system_global == listener->mode->global))
@@ -1218,18 +1335,19 @@ spi_keystroke_from_x_key_event (XKeyEvent *x_key_event)
   Accessibility_DeviceEvent key_event;
   KeySym keysym;
   const int cbuf_bytes = 20;
-  char cbuf [cbuf_bytes];
-  
-  keysym = XLookupKeysym (x_key_event, 0);
+  char cbuf [cbuf_bytes+1];
+  int nbytes;
+
+  nbytes = XLookupString (x_key_event, cbuf, cbuf_bytes, &keysym, NULL);  
   key_event.id = (CORBA_long)(keysym);
   key_event.hw_code = (CORBA_short) x_key_event->keycode;
   if (((XEvent *) x_key_event)->type == KeyPress)
     {
-      key_event.type = Accessibility_KEY_PRESSED;
+      key_event.type = Accessibility_KEY_PRESSED_EVENT;
     }
   else
     {
-      key_event.type = Accessibility_KEY_RELEASED;
+      key_event.type = Accessibility_KEY_RELEASED_EVENT;
     } 
   key_event.modifiers = (CORBA_unsigned_short)(x_key_event->state);
   key_event.is_text = CORBA_FALSE;
@@ -1239,9 +1357,6 @@ spi_keystroke_from_x_key_event (XKeyEvent *x_key_event)
         key_event.event_string = CORBA_string_dup ("space");
         break;
       case XK_Tab:
-#ifdef SPI_KEYEVENT_DEBUG
-	fprintf(stderr, "Tab\n");
-#endif
         key_event.event_string = CORBA_string_dup ("Tab");
 	break;
       case XK_BackSpace:
@@ -1314,12 +1429,16 @@ spi_keystroke_from_x_key_event (XKeyEvent *x_key_event)
         key_event.event_string = CORBA_string_dup ("Right");
 	break;
       default:
-        if (XLookupString (x_key_event, cbuf, cbuf_bytes, &keysym, NULL) > 0)
+        if (nbytes > 0)
           {
+	    gunichar c;
+	    cbuf[nbytes] = '\0'; /* OK since length is cbuf_bytes+1 */
             key_event.event_string = CORBA_string_dup (cbuf);
-	    if (isgraph (keysym))
+	    c = g_utf8_get_char_validated (cbuf, nbytes);
+	    if ((c > 0) && g_unichar_isprint (c))
 	      {
-	        key_event.is_text = CORBA_TRUE; /* FIXME: incorrect for some composed chars? */
+	        key_event.is_text = CORBA_TRUE; 
+		/* incorrect for some composed chars? */
 	      }
           }
         else
@@ -1330,14 +1449,29 @@ spi_keystroke_from_x_key_event (XKeyEvent *x_key_event)
 
   key_event.timestamp = (CORBA_unsigned_long) x_key_event->time;
 #ifdef SPI_KEYEVENT_DEBUG
-  fprintf (stderr,
-     "Key %lu pressed (%c), modifiers %d\n",
-     (unsigned long) keysym,
-     keysym ? (int) keysym : '*',
-     (int) x_key_event->state);
+  {
+    char *pressed_str  = "pressed";
+    char *released_str = "released";
+    char *state_ptr;
+
+    if (key_event.type == Accessibility_KEY_PRESSED_EVENT)
+      state_ptr = pressed_str;
+    else
+      state_ptr = released_str;
+ 
+    fprintf (stderr,
+	     "Key %lu %s (%c), modifiers %d; string=%s [%x] %s\n",
+	     (unsigned long) keysym,
+	     state_ptr,
+	     keysym ? (int) keysym : '*',
+	     (int) x_key_event->state,
+	     key_event.event_string,
+	     key_event.event_string[0],
+	     (key_event.is_text == CORBA_TRUE) ? "(text)" : "(not text)");
+  }
 #endif
 #ifdef SPI_DEBUG
-  fprintf (stderr, "%s%c",
+  fprintf (stderr, "%s%c\n",
      (x_key_event->state & Mod1Mask)?"Alt-":"",
      ((x_key_event->state & ShiftMask)^(x_key_event->state & LockMask))?
      g_ascii_toupper (keysym) : g_ascii_tolower (keysym));
@@ -1372,9 +1506,8 @@ spi_controller_update_key_grabs (SpiDEController           *controller,
       next = l->next;
 
       re_issue_grab = recv &&
-/*	      (recv->type == Accessibility_KEY_RELEASED) && - (?) */
 	      (recv->modifiers & grab_mask->mod_mask) &&
-	      (grab_mask->key_val == keycode_for_keysym (recv->id));
+	      (grab_mask->key_val == keycode_for_keysym (recv->id, NULL));
 
 #ifdef SPI_DEBUG
       fprintf (stderr, "mask=%lx %lx (%c%c) %s\n",
@@ -1458,6 +1591,16 @@ spi_device_event_controller_object_finalize (GObject *object)
   /* disconnect any special listeners, get rid of outstanding keygrabs */
   XUngrabKey (spi_get_display (), AnyKey, AnyModifier, DefaultRootWindow (spi_get_display ()));
 
+#ifdef HAVE_XEVIE
+  if (controller->xevie_display != NULL)
+    {
+      XevieEnd(controller->xevie_display);
+#ifdef SPI_KEYEVENT_DEBUG
+      printf("XevieEnd(dpy) finished \n");
+#endif
+    }
+#endif
+
   private = g_object_get_data (G_OBJECT (controller), "spi-dec-private");
   if (private->xkb_desc)
 	  XkbFreeKeyboard (private->xkb_desc, 0, True);
@@ -1782,6 +1925,158 @@ dec_synth_keycode_release (SpiDEController *controller,
 	return TRUE;
 }
 
+static unsigned
+dec_get_modifier_state (SpiDEController *controller)
+{
+	return mouse_mask_state;
+}
+
+static gboolean
+dec_lock_modifiers (SpiDEController *controller, unsigned modifiers)
+{
+	return XkbLockModifiers (spi_get_display (), XkbUseCoreKbd, 
+			  modifiers, modifiers);
+}
+
+static gboolean
+dec_unlock_modifiers (SpiDEController *controller, unsigned modifiers)
+{
+	return XkbLockModifiers (spi_get_display (), XkbUseCoreKbd, 
+			  modifiers, 0);
+}
+
+static KeySym
+dec_keysym_for_unichar (SpiDEController *controller, gunichar unichar)
+{
+	/* TODO: table lookups within a range, for various code pages! */
+	KeySym keysym = NoSymbol;
+
+	if (unichar >= 0x20 && unichar < 0x7f) { /* Basic Latin/ASCII */
+		keysym = (KeySym) unichar;
+	}
+	else if (unichar >= 0xa0 && unichar <= 0xff) { /* Latin 1 extensions */
+		keysym = (KeySym) unichar;
+	}
+	else if (unichar >= 0x100 && unichar <= 0x233) { /* unfortunately the mapping gets nasty for Latin-2 and 3... help! */
+		keysym = NoSymbol;
+	}
+	else if (unichar >= 0x7c1 && unichar <= 0x3a1) { /* let's try Greek anyway... */
+		keysym = (KeySym) (0x391 + (unichar - 0x7c1));
+	}
+	else if (unichar >= 0x3a3 && unichar <= 0x3a9) { /* let's try Greek anyway... */
+		keysym = (KeySym) (0x7d2 + (unichar - 0x3a3));
+	}
+	else if (unichar >= 0x3b1 && unichar <= 0x3c1) { /* let's try Greek anyway... */
+		keysym = (KeySym) (0x7e1 + (unichar - 0x3b1));
+	}
+	else if (unichar == 0x3c2) {
+		keysym = (KeySym) 0x7f3; /* XK_Greek_finalsmallsigma; */
+	}
+	else if (unichar >= 0x3c3 && unichar <= 0x3c9) { /* let's try Greek anyway... */
+		keysym = (KeySym) (0x7f2 + (unichar - 0x3c2));
+	}	
+	else if (unichar >= 0x5d0 && unichar <= 0x5ea) { /* Hebrew basics */
+		/* probably broken :-) */
+		keysym = (KeySym) (0xce0 + (unichar - 0x5d0));
+	}	
+	else if (unichar >= 0x30a1 && unichar <= 0x30ab) { /* partial katakana support */
+		/* TODO: complete! */
+		keysym = (KeySym) (0x4b1 + (unichar - 0x30a1)/2);
+	}
+	else if (unichar >= 0x20a0 && unichar <= 0x20ac) { /* currency */
+		keysym = (KeySym) unichar; /* how convenient ;-) */
+	}
+	return keysym;
+}
+
+static gboolean
+dec_synth_keysym (SpiDEController *controller, KeySym keysym)
+{
+	KeyCode key_synth_code;
+	unsigned int modifiers, synth_mods, lock_mods;
+
+	key_synth_code = keycode_for_keysym (keysym, &synth_mods);
+
+	if ((key_synth_code == 0) || (synth_mods == 0xFF)) return FALSE;
+
+	/* TODO: set the modifiers accordingly! */
+	modifiers = dec_get_modifier_state (controller);
+	/* side-effect; we may unset mousebutton modifiers here! */
+
+	if (synth_mods != modifiers) {
+		lock_mods = synth_mods & ~modifiers;
+		dec_lock_modifiers (controller, lock_mods);
+	}
+	dec_synth_keycode_press (controller, key_synth_code);
+	dec_synth_keycode_release (controller, key_synth_code);
+	if (synth_mods != modifiers) 
+		dec_unlock_modifiers (controller, lock_mods);
+	return TRUE;
+}
+
+
+static gboolean
+dec_synth_keystring (SpiDEController *controller, const CORBA_char *keystring)
+{
+	/* probably we need to create and inject an XIM handler eventually. */
+	/* for now, try to match the string to existing 
+	 * keycode+modifier states. 
+         */
+	KeySym *keysyms;
+	gint maxlen = 0;
+	gunichar unichar = 0;
+	gint i = 0;
+	gboolean retval = TRUE;
+	const gchar *c;
+
+	maxlen = strlen (keystring);
+	keysyms = g_new0 (KeySym, maxlen);
+	if (!(keystring && *keystring && g_utf8_validate (keystring, -1, &c))) { 
+		retval = FALSE;
+	} 
+	else {
+#ifdef SPI_DEBUG
+		fprintf (stderr, "[keystring synthesis attempted on %s]\n", keystring);
+#endif
+		while (keystring && (unichar = g_utf8_get_char (keystring))) {
+			KeySym keysym;
+			char bytes[6];
+			gint mbytes;
+			
+			mbytes = g_unichar_to_utf8 (unichar, bytes);
+			bytes[mbytes] = '\0';
+#ifdef SPI_DEBUG
+		        fprintf (stderr, "[unichar %s]", bytes);
+#endif
+			keysym = dec_keysym_for_unichar (controller, unichar);
+			if (keysym == NoSymbol) {
+#ifdef SPI_DEBUG
+				fprintf (stderr, "no keysym for %s", bytes);
+#endif
+				retval = FALSE;
+				break;
+			}
+			keysyms[i++] = keysym;
+			keystring = g_utf8_next_char (keystring); 
+		}
+		keysyms[i++] = 0;
+		for (i = 0; keysyms[i]; ++i) {
+			if (!dec_synth_keysym (controller, keysyms[i])) {
+#ifdef SPI_DEBUG
+				fprintf (stderr, "could not synthesize %c\n",
+					 (int) keysyms[i]);
+#endif
+				retval = FALSE;
+				break;
+			}
+		}
+	}
+	g_free (keysyms);
+
+	return retval;
+}
+
+
 /*
  * CORBA Accessibility::DEController::registerKeystrokeListener
  *     method implementation
@@ -1795,11 +2090,8 @@ impl_generate_keyboard_event (PortableServer_Servant           servant,
 {
   SpiDEController *controller =
 	SPI_DEVICE_EVENT_CONTROLLER (bonobo_object (servant));
-  DEControllerPrivateData *priv;
   long key_synth_code;
-  unsigned int slow_keys_delay;
-  unsigned int press_time;
-  unsigned int release_time;
+  KeySym keysym;
 
 #ifdef SPI_DEBUG
 	fprintf (stderr, "synthesizing keystroke %ld, type %d\n",
@@ -1812,7 +2104,6 @@ impl_generate_keyboard_event (PortableServer_Servant           servant,
    * and fall back to XSendEvent() if XTest is not available.
    */
   
-  /* TODO: implement keystring mode also */
   gdk_error_trap_push ();
   key_synth_code = keycode;
 
@@ -1830,18 +2121,33 @@ impl_generate_keyboard_event (PortableServer_Servant           servant,
 #ifdef SPI_XKB_DEBUG	      
 	      fprintf (stderr, "KeySym synthesis\n");
 #endif
-	      key_synth_code = keycode_for_keysym (keycode);
-	      dec_synth_keycode_press (controller, key_synth_code);
-	      dec_synth_keycode_release (controller, key_synth_code);
+	      /* 
+	       * note: we are using long for 'keycode'
+	       * in our arg list; it can contain either
+	       * a keycode or a keysym.
+	       */
+	      dec_synth_keysym (controller, (KeySym) keycode);
 	      break;
       case Accessibility_KEY_STRING:
-	      fprintf (stderr, "Not yet implemented\n");
+	      if (!dec_synth_keystring (controller, keystring))
+		      fprintf (stderr, "Keystring synthesis failure, string=%s\n",
+			       keystring);
 	      break;
     }
   if (gdk_error_trap_pop ())
     {
       DBG (-1, g_warning ("Error emitting keystroke"));
     }
+  if (synth_type == Accessibility_KEY_SYM) {
+    keysym = keycode;
+  }
+  else {
+    keysym = XkbKeycodeToKeysym (spi_get_display (), keycode, 0, 0);
+  }
+  if (XkbKeysymToModifiers (spi_get_display (), keysym) == 0) 
+    {
+      spi_dec_clear_unlatch_pending (controller);
+    }
 }
 
 /* Accessibility::DEController::generateMouseEvent */
@@ -1852,7 +2158,7 @@ impl_generate_mouse_event (PortableServer_Servant servant,
 			   const CORBA_char      *eventName,
 			   CORBA_Environment     *ev)
 {
-  int button;
+  int button = 0;
   gboolean error = FALSE;
   Display *display = spi_get_display ();
 #ifdef SPI_DEBUG
@@ -1874,6 +2180,12 @@ impl_generate_mouse_event (PortableServer_Servant servant,
 	  case '3':
 	          button = 3;
 	          break;
+	  case '4':
+	          button = 4;
+	          break;
+	  case '5':
+	          button = 5;
+	          break;
 	  default:
 		  error = TRUE;
 	  }
@@ -1958,21 +2270,75 @@ spi_device_event_controller_class_init (SpiDEControllerClass *klass)
 	  spi_dec_private_quark = g_quark_from_static_string ("spi-dec-private");
 }
 
+#ifdef HAVE_XEVIE
+Bool isEvent(dpy,event,arg)
+     Display *dpy;
+     XEvent *event;
+     char *arg;
+{
+   return TRUE;
+}
+
+gboolean
+handle_io (GIOChannel *source,
+           GIOCondition condition,
+           gpointer data) 
+{
+  SpiDEController *controller = (SpiDEController *) data;
+  gboolean is_consumed = FALSE;
+  XEvent ev;
+
+  while (XCheckIfEvent(controller->xevie_display, &ev, isEvent, NULL))
+    {
+      if (ev.type == KeyPress || ev.type == KeyRelease)
+        is_consumed = spi_device_event_controller_forward_key_event (controller, &ev);
+
+      if (! is_consumed)
+        XevieSendEvent(controller->xevie_display, &ev, XEVIE_UNMODIFIED);
+    }
+
+  return TRUE;
+}
+#endif /* HAVE_XEVIE */
+
 static void
 spi_device_event_controller_init (SpiDEController *device_event_controller)
 {
+#ifdef HAVE_XEVIE
+  GIOChannel *ioc;
+  int fd;
+#endif /* HAVE_XEVIE */
+
   DEControllerPrivateData *private;	
   device_event_controller->key_listeners   = NULL;
   device_event_controller->mouse_listeners = NULL;
   device_event_controller->keygrabs_list   = NULL;
+  device_event_controller->xevie_display   = NULL;
+
+#ifdef HAVE_XEVIE
+  device_event_controller->xevie_display = XOpenDisplay(NULL);
+
+  if (XevieStart(device_event_controller->xevie_display) == TRUE)
+    {
+#ifdef SPI_KEYEVENT_DEBUG
+      fprintf (stderr, "XevieStart() success \n");
+#endif
+      XevieSelectInput(device_event_controller->xevie_display, KeyPressMask | KeyReleaseMask);
+
+      fd = ConnectionNumber(device_event_controller->xevie_display);
+      ioc = g_io_channel_unix_new (fd);
+      g_io_add_watch (ioc, G_IO_IN | G_IO_HUP, handle_io, device_event_controller);
+      g_io_channel_unref (ioc);
+    }
+  else
+    {
+      device_event_controller->xevie_display = NULL;
+#ifdef SPI_KEYEVENT_DEBUG
+      fprintf (stderr, "XevieStart() failed, only one client is allowed to do event int exception\n");
+#endif
+    }
+#endif /* HAVE_XEVIE */
 
-  /*
-   * TODO: fixme, this module makes the foolish assumptions that
-   * registryd uses the same display as the apps, and that the
-   * DISPLAY environment variable is set.
-   */
-  gdk_init (NULL, NULL);
-  
   private = g_new0 (DEControllerPrivateData, 1);
   gettimeofday (&private->last_press_time, NULL);
   gettimeofday (&private->last_release_time, NULL);
@@ -1982,11 +2348,10 @@ spi_device_event_controller_init (SpiDEController *device_event_controller)
   spi_controller_register_with_devices (device_event_controller);
 }
 
-static void
+static gboolean
 spi_device_event_controller_forward_key_event (SpiDEController *controller,
 					       const XEvent    *event)
 {
-  gboolean is_consumed = FALSE;
   CORBA_Environment ev;
   Accessibility_DeviceEvent key_event;
 
@@ -1996,20 +2361,11 @@ spi_device_event_controller_forward_key_event (SpiDEController *controller,
 
   key_event = spi_keystroke_from_x_key_event ((XKeyEvent *) event);
 
-  spi_controller_update_key_grabs (controller, &key_event);
+  if (controller->xevie_display == NULL)
+    spi_controller_update_key_grabs (controller, &key_event);
 
   /* relay to listeners, and decide whether to consume it or not */
-  is_consumed = spi_controller_notify_keylisteners (
-	  controller, &key_event, CORBA_TRUE, &ev);
-
-  if (is_consumed)
-    {
-      XAllowEvents (spi_get_display (), AsyncKeyboard, CurrentTime);
-    }
-  else
-    {
-      XAllowEvents (spi_get_display (), ReplayKeyboard, CurrentTime);
-    }
+  return spi_controller_notify_keylisteners (controller, &key_event, CORBA_TRUE, &ev);
 }
 
 SpiDEController *
@@ -2017,7 +2373,6 @@ spi_device_event_controller_new (SpiRegistry *registry)
 {
   SpiDEController *retval = g_object_new (
     SPI_DEVICE_EVENT_CONTROLLER_TYPE, NULL);
-  DEControllerPrivateData *private;
   
   retval->registry = SPI_REGISTRY (bonobo_object_ref (
 	  BONOBO_OBJECT (registry)));
@@ -2030,4 +2385,4 @@ spi_device_event_controller_new (SpiRegistry *registry)
 BONOBO_TYPE_FUNC_FULL (SpiDEController,
 		       Accessibility_DeviceEventController,
 		       PARENT_TYPE,
-		       spi_device_event_controller);
+		       spi_device_event_controller)
