@@ -41,6 +41,7 @@ static void validate_accessible (Accessible *accessible,
 
 #define WINDOW_MAGIC 0x123456a
 #define TEST_STRING_A "A test string"
+#define TEST_STRING_A_OBJECT "A_test_string_object"
 #define TEST_STRING_B "Another test string"
 
 static int      print_tree_depth = 0;
@@ -105,6 +106,7 @@ create_test_window (void)
 {
 	TestWindow *win = g_new0 (TestWindow, 1);
 	GtkWidget  *widget, *vbox;
+	AtkObject *obj;
 
 	win->magic  = WINDOW_MAGIC;
 	win->window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
@@ -117,6 +119,9 @@ create_test_window (void)
 
 	widget = gtk_entry_new ();
 	gtk_entry_set_text (GTK_ENTRY (widget), TEST_STRING_A);
+	obj = gtk_widget_get_accessible (widget);
+	atk_object_set_name (obj, TEST_STRING_A_OBJECT);
+
 	test_window_add_and_show (GTK_CONTAINER (vbox), widget);
 
 	widget = gtk_button_new_with_label ("_Foobar");
@@ -129,7 +134,7 @@ create_test_window (void)
 					   GTK_ICON_SIZE_LARGE_TOOLBAR);
 	test_window_add_and_show (GTK_CONTAINER (vbox), widget);
 
-	widget = g_object_new (GTK_TYPE_RANGE, NULL);
+	widget = g_object_new (GTK_TYPE_HSCALE, NULL);
 	gtk_range_set_range (GTK_RANGE (widget), 0.0, 100.0);
 	test_window_add_and_show (GTK_CONTAINER (vbox), widget);
 
@@ -157,8 +162,8 @@ test_roles (void)
 	for (i = -1; i < 1000; i++)
 		g_assert (AccessibleRole_getName (i) != NULL);
 
-	g_assert (!strcmp (AccessibleRole_getName (SPI_ROLE_FILE_CHOOSER), "file chooser"));
-	g_assert (!strcmp (AccessibleRole_getName (SPI_ROLE_RADIO_BUTTON), "radiobutton"));
+	g_assert (!strcmp (AccessibleRole_getName (SPI_ROLE_FILE_CHOOSER), "file-chooser"));
+	g_assert (!strcmp (AccessibleRole_getName (SPI_ROLE_RADIO_BUTTON), "radio-button"));
 	g_assert (!strcmp (AccessibleRole_getName (SPI_ROLE_TABLE), "table"));
 	g_assert (!strcmp (AccessibleRole_getName (SPI_ROLE_WINDOW), "window"));
 }
@@ -600,8 +605,10 @@ validate_accessible (Accessible *accessible,
 		g_assert (tmp != NULL);
 		if (print_tree)
 			fprintf (stderr, "Te");
-		else
-			test_text (tmp);
+		else {
+			if (strcmp (name, TEST_STRING_A_OBJECT) == 0)	
+				test_text (tmp);
+		}
 		AccessibleText_unref (tmp);
 	}
 
@@ -692,7 +699,6 @@ key_listener_cb (const AccessibleKeystroke *stroke,
 static void
 test_keylisteners (void)
 {
-#ifdef BILL_MAKES_THIS_WORK_RELIABLY
 	int i;
 	AccessibleKeystroke stroke;
 	AccessibleKeystrokeListener *key_listener;
@@ -704,7 +710,7 @@ test_keylisteners (void)
 		key_listener_cb, &stroke);
 
 	test_keyset = SPI_createAccessibleKeySet (1, "=", NULL, NULL);
-	
+
 	g_assert (SPI_registerAccessibleKeystrokeListener (
 		key_listener,
 		test_keyset,
@@ -736,7 +742,6 @@ test_keylisteners (void)
         g_assert (SPI_generateMouseEvent (-50, -50, "rel"));		  
         g_assert (SPI_generateMouseEvent (-50, -50, "rel"));		  
         g_assert (SPI_generateMouseEvent (-1, -1, "b1c")); 
-#endif
 }
 
 int
@@ -771,6 +776,7 @@ main (int argc, char **argv)
 	win = create_test_window ();
 
 	global_listener = SPI_createAccessibleEventListener (global_listener_cb, win);
+
 	g_assert (SPI_registerGlobalEventListener (global_listener, "focus:"));
 
 	fprintf (stderr, "Waiting for focus event ...\n");
@@ -784,7 +790,7 @@ main (int argc, char **argv)
 	/* Wait for any pending events from the registry */
 	g_usleep (500*1000);
 	for (i = 0; i < 100; i++)
-		linc_main_iteration (FALSE);
+		CORBA_ORB_perform_work (NULL, NULL);
 
 	if ((leaked = SPI_exit ()))
 		g_error ("Leaked %d SPI handles", leaked);
